;; Copyright (c) 1994-1998 Franz Inc, Berkeley, Ca.
;;
;; Permission is granted to any individual or institution to use, copy,
;; modify, and distribute this software, provided that this complete
;; copyright and permission notice is maintained, intact, in all copies and
;; supporting documentation.
;;
;; Franz Incorporated provides this software "as is" without
;; express or implied warranty.


;; Utility to scan Lucid Common Lisp source files
;; for package qualifiers that might cause trouble
;; when porting to Allegro CL.
;;
;; The main (and only) function is scan-file-for-qualifiers.

(defvar scan-file-marker "??"
  "A marker inserted in a file to flag a package qualifier problem.")

(defvar scan-file-lucid-package-names
    '("LISP"
      "CL"
      "COMMON-LISP"
      "CLOS"
      "LUCID-COMMON-LISP" "LCL"     
      
      "CLOS-SYSTEM"
      "EPITOME"
      "FLAVORS"
      "LOOP"
      "LUCID"
      "LUCID-RUNTIME-SUPPORT"  "LRS"
      "LWALKER"
      "MONITOR"
      "PRETTY-PRINTER"  "XP"  "PP"
      "PSEUDO-CLASS"
      "REORGANIZER"
      "SETF"
      "SNOOPY"
      "STATE-STEP"
      "SYSTEM" "SYS"
      "TREESHAKE"
      "UDEFSYS"
      "WINDOWS"
      "XLT"
      "XLT/CLX"
      "XTOOLS"
      ))


(defun scan-file-for-qualifiers ()
  "Search the entire buffer for package qualifiers that might cause
problems when the file is read by Allegro CL.

When one is found, prompt for an action:

 Keep [Space] - press the space bar to keep the buffer as is
 Del [d, k]   - press d or k to delete the qualifier entirely
 lcl: [l]     - press l to replace the qualifier with lcl:
 ?: [m]       - press m to replace the qualifier with a marker
 Quit [q]     - press q to stop the scan at the current place

The marker is defined by the value of the variable scan-file-marker.

At each prompt, point is positioned at the BEGINNING of the suspected
package qualifier which extends to the first or second : to the right.
Thus, the qualified name mlcl:foo will cause two (bogus) prompts, one
for the qualifier cl: and one for lcl:."
  (interactive)
  (let ((begin (point)) (top (point-min)) done)
    (block scan-file
      (goto-char top)
      (if (not (search-forward ":" nil t))
	  nil
	(do ((pkl scan-file-lucid-package-names (cdr pkl)) pk)
	    ((atom pkl))
	  (goto-char top)
	  (setq pk (car pkl))
	  (loop
	   (if (null (search-forward-regexp (concat (downcase pk) ":[^ )]") nil t))
	       (return)			; continue the search with another qualifier
	     ;; Otherwise, try to suggest some remedial action:
	     ;; LCL::xx   could be dropped or changed to LCL:
	     ;; LCL:xx    could be dropped
	     ;; SYS::xx   could be dropped or changed to LCL:
	     ;; xxx:[:] needs to be handled
	     (let ((end (- (point) 1))
		   (beg (- (point) (length pk) 2))
		   mid cmd)
	       (if done nil (setq done 0))
	       ;; move point to non-blank char after :
	       (goto-char (- (point) 1))
	       ;; remember position of first :
	       (setq mid (- (point) 1))
	       (if (looking-at ":")
		   ;; cover entire qualifier of internal symbol
		   (setq end (+ end 1)))
	       (goto-char beg)
	       (message
		"Package qual? Keep [Space]; Del [d,k]; lcl: [l]; Mark [m]; Quit [q]")
	       (setq cmd (read-char))
	       (case cmd 
		 ((?d ?D ?k ?K) (setq cmd 'delete))
		 ((?l ?L) (setq cmd 'change))
		 ((?m ?M) (setq cmd 'mark))
		 ((?q ?Q) (setq cmd 'quit))
		 (otherwise (setq cmd 'keep)))
	       (case cmd
		 (keep (setq done (- done 1)))
		 (delete (delete-char (- end beg)))
		 (change (delete-char (- end beg))
			 (insert "lcl:"))
		 (mark   (goto-char mid)
			 (insert scan-file-marker))
		 (quit (return-from scan-file nil))
		 )
	       (setq done (+ done 1))
	       (goto-char end)
	       ))
	   )
	  ))
      (goto-char begin)
      )
    (case done
	((nil) (message
		"There are no package qualifiers at all in this buffer."))
	(0 (message
	    "No package qualifiers were changed in this buffer."))
	(1 (message
	    "One package qualifier was changed in this buffer."))
	(otherwise (message
		    "%d package qualifiers were changed in this buffer." done))
	)))
