;;
;; ftp client for lisp
;;
;; copyright (c) 1996-1998 Franz Inc, Berkeley, CA  - All rights reserved.
;;
;; This file contains two exported functions:
;;	open-ftp-stream -- open a stream to read/write a file via ftp
;; 	ftp-transfer-file -- transfer a file between two machines
;;
;; The documentation for these functions appears with the function
;; definition itself.
;;
;;
;; [see rfc959 for the specification of the ftp protocol]
;;
;; $Id: ftp.cl,v 1.4 1998/03/17 23:30:42 duane Exp $
;;
;;
(defpackage :socket
  (:use #+aclpc :lisp #-aclpc :common-lisp :clos :excl)
  (:export #:open-ftp-stream
	   #:ftp-transfer-file))

(in-package :socket)



;; response-class gets a response from the ftp server and
;; dispatches on that response.  The variable ftp-response
;; is bound to the complete response string from the server
;;
;; use:
;;   (response-case (stream)
;;     (2
;;       (action-for-2xx-response))
;;     (3
;;	 (action-for-3xx-response))
;;     (t (error "bogus response ~s" ftp-response)))
;;
(defmacro response-case ((ftp-stream) &rest case-clauses)
  (let ((response-class (gensym)))
    `(multiple-value-bind (,response-class ftp-response)
	 (wait-for-response ,ftp-stream)
       (case ,response-class
	 ,@case-clauses))))


(defvar *ftp-debug* nil)


;;; exported functions:


(defun open-ftp-stream (file &key
			     (host "localhost")
			     (user "anonymous")
			     (password "hacker@lisp.programer")
			     (direction	:input) ; :input or :output
			     debug	        ; t or stream
			     (mode :binary)       ; :text or :binary
			     (passive nil)	; t if use passive mode
			     )

  ;; create a stream to/from the given file on the given host
  ;; using the given user and password for authentication.

  ;;  keyword arguments:
  ;;    :host  - name of remote host (default is localhost)
  ;;    :user  - name of user to log in as (default in anonymous)
  ;;    :password - password to use  (default is bogus string)
  ;;    :direction - :input (default) or :output
  ;;    :debug  - if a stream (or t) is given it will print the ftp
  ;;		interaction to that stream
  ;;    :mode - :binary (default) or :text.  Note that in :text mode
  ;;	    lines end with cr-lf (the internet standard)
  ;;    :passive - if t then do a passive mode transfer (useful to get
  ;; 	    through certain ip filters and firewalls).


  (let* ((*ftp-debug* debug)
	 (ftpd (connect-to-ftp host user password)))
    (unwind-protect
	(if* passive
	   then (passive-ftp-stream ftpd file direction mode)
	   else (active-ftp-stream ftpd file direction mode))
      (close ftpd))))


(defun ftp-transfer-file
    (&key from-host from-file
	  (from-user "anonymous")
	  (from-password "acl@franz.lisp")

	  to-host to-file
	  (to-user "anonymous")
	  (to-password "acl@franz.lisp")

	  (mode :binary)
	  debug
	  )

  ;; transfer the file from the 'from' machine to the 'to' machine.
  ;; the files on the different machines can have different names
  ;;
  ;; This is an implementation of a 3-machine ftp that was
  ;; mentioned as being possible in the rfc but I've never
  ;; seen an ftp client implement it ... until now.

  (let (to-ftpd from-ftpd (*ftp-debug* debug))
    (unwind-protect
	(progn
	  (setq to-ftpd (connect-to-ftp to-host to-user to-password)
		from-ftpd (connect-to-ftp from-host from-user from-password))

	  ;; set them both in the mode
	  (ftp-type-command to-ftpd mode)
	  (ftp-type-command from-ftpd mode)

	  ;; make the 'to' machine be passive
	  (multiple-value-bind (to-ipaddr to-port)
	      (ftp-pasv-command to-ftpd)

	    ;; tell the from machine where to contact the to machine.
	    (ftp-port-command from-ftpd to-ipaddr to-port)

	    ;; start the transfer on each side
	    (ftp-transfer-command from-ftpd
				  from-file
				  :input)

	    (ftp-transfer-command  to-ftpd
				   to-file
				   :output)

	    (response-case (from-ftpd)
			   (2 ;; ok
			    nil)
			   (t (error "ftp from failed: ~a" ftp-response)))
	    (response-case (to-ftpd)
			   (2 ;; ok
			    nil)
			   (t (error "ftp to failed: ~a" ftp-response)))
	    nil
	    ))
      ;; cleanup forms:
      (if* from-ftpd then (close from-ftpd))
      (if* to-ftpd   then (close to-ftpd)))))


;;;;;;;;;;;;;  end exported functions



(defun connect-to-ftp (host user password &key (port "ftp"))
  ;; connect to the ftp server on 'host'
  ;; login as user with the given password
  ;; return a stream for talking to the ftp server, signal an error
  ;;  and close the stream if there is a problem
  (let ((ftpd (make-socket :remote-host host
			   :remote-port port))
	(ok nil))
    (unwind-protect
	(progn
	  ;; check initial message from ftp server..
	  ;; It may be too busy to talk to us
	  (response-case (ftpd)
			 (2 ;; success
			  nil)
			 (t (error "ftp ~s responded ~a"
				   host
				   ftp-response)))


	  (ftp-command ftpd "user ~a" user)
	  (response-case
	   (ftpd)
	   (2 ;; we're in with no password
	    nil)
	   (3 ;; need password
	    (ftp-command ftpd "pass ~a" password)
	    (response-case (ftpd)
			   (2 ;; we're in
			    nil)
			   (3 ;; need acct,,, we don't know what that is
			    (error "ftp daemon needs acct info"))
			   (t (error "ftp ~s doesn't like the password: ~a"
				     host
				     ftp-response
				     ))))
	   (t (error "ftp ~s doesn't like the user name: ~a: ~a"
		     host
		     user
		     ftp-response)))
	  (setq ok t)	;; all is well.. do don't close the port
	  ftpd)
      (if* (not ok) then (close ftpd))
      )))



(defun active-ftp-stream (ftpd file direction mode)
  ;; this is the normal (non-passive) way ftp works:
  ;; the client announces a port and the server connects to it regardless
  ;; of which way the data is travelling.
  ;; The problem with this is that some companies have gateways and
  ;; firewalls that prevent outsiders from connecting to ports on machines
  ;; on the inside.  Thus there is a passive mode (see passive-ftp-stream)
  ;;
  ;; This function sets up a stream to read or write the remote file.
  ;; direction is :input or :output
  ;; mode is :text or :binary
  ;;
  (let ((datap (make-socket :connect :passive)))
    (unwind-protect
	(progn

	  ;; tell the remote site about the our port
	  ;; Since the local-host of a passive socket (e.g. datap)
	  ;; has no useful info we'll use the local-host value from ftpd.
	  ;;

	  (ftp-port-command ftpd (local-host ftpd) (local-port datap))

	  ;; specify the type
	  (ftp-type-command ftpd mode)

	  ;; now retrieve or store the file
	  (ftp-transfer-command ftpd file direction)

	  ;; ftp daemon should be connecting to us
	  ;; we'll accept the connection and then close off the passive
	  ;; socket, leaving only the socket to ftp's control connection
	  ;; up and running
	  (accept-connection datap))
      (close datap))))


(defun passive-ftp-stream (ftpd file direction mode)
  ;; do a passive mode file transfer
  ;; In this mode the server announces a port and we connect to it
  ;;
  (unwind-protect
      (multiple-value-bind (rem-ipaddr rem-port)
	  (ftp-pasv-command ftpd)

	;; specify the file type
	(ftp-type-command ftpd mode)
	(ftp-transfer-command ftpd file direction
			      #'(lambda ()
				  (make-socket :remote-host rem-ipaddr
					       :remote-port rem-port)))

	)
    (close ftpd)))


(defun ftp-port-command (ftpd ipaddr port)
  ;; specify a port on which the remote machine will contact us
  (ftp-command ftpd
	       "port ~d,~d,~d,~d,~d,~d"
	       (logand #xff (ash ipaddr -24))
	       (logand #xff (ash ipaddr -16))
	       (logand #xff (ash ipaddr -8))
	       (logand #xff ipaddr)

	       (logand #xff (ash port -8))
	       (logand #xff port))
  (response-case (ftpd)
		 (2 ;; ok!
		  nil
		  )
		 (t (error "ftpd server didn't like port command: ~a"
			   ftp-response))))



(defun ftp-pasv-command (ftpd)
  ;; issue a 'passive' command and return two values:
  ;; the ipaddr and port on the remote host we should contact
  (ftp-command ftpd "pasv")
  (response-case (ftpd)
		 (2 (parse-ipaddr-port ftp-response))
		 (t (error "ftp doesn't like the pasv command: ~s"
			   ftp-response))))

(defun parse-ipaddr-port (str)
  ;; the string has the form  "..... (h1,h2.h3.h4,p1.p2)"
  ;; where h1 is the upper byte of the ipaddr and so on.
  ;; return two values: the ipaddr and the port
  ;;
  (let ((offset (position #\( str))
	(max (length str))
	(res nil))
    (do ((i (1+ offset) (1+ i))
	 (accum 0)
	 (ch))
	((>= i max))
      (setq ch (aref str i))
      (let ((v (digit-char-p ch)))
	(if* v
	   then (setq accum (+ (* accum 10) v))
	 elseif (eql ch #\,)
	   then (push accum res)
		(setq accum 0)
	 elseif (eql ch #\))
	   then (push accum res)
		(return)	; all done
		)))
    (if* (not (eql (length res) 6))
       then (error "can't find the ipaddr and port in this response: ~s"
		   str))
    ;; at this point res is reversed
    (values
     (+ (nth 2 res)
	(ash (nth 3 res) 8)
	(ash (nth 4 res) 16)
	(ash (nth 5 res) 24))
     (+ (nth 0 res)
	(ash (nth 1 res) 8)))))






(defun ftp-type-command (ftpd mode)
  ;;
  ;; issue the ftp command to set type of the transfer
  ;;
  (ftp-command ftpd "type ~a" (ecase mode
				(:text "a")
				(:binary "i")))
  (response-case (ftpd)
		 (2 ;; ok
		  nil)
		 (t (error "ftp server's response to type command: ~a"
			   ftp-response))))

(defun ftp-transfer-command (ftpd file direction &optional action)
  ;;
  ;; issue the command to initiate the transfer.
  ;; optionally run an action between the sending of the command
  ;; and the checking of the result code.  This is necessary to
  ;; do a passive transfer since the control connnection wants to see
  ;; the connect before it will issue a result code.
  ;;
  (ftp-command ftpd "~a ~a"
	       (ecase direction
		 (:input "retr")
		 (:output "stor"))
	       file)
  (prog1 (if* action
	    then (funcall action))

    (response-case (ftpd)
		   ((1 2)
		    ;; transfer about to start
		    nil
		    )
		   (t (error "ftp transfer command failed: ~a"
			     ftp-response)))
    ))



(defun wait-for-response (stream)
  ;; read the response of the ftp server.
  ;; collect it all in a string.
  ;; Return two values:
  ;; 	response class
  ;;    whole string
  ;; The string should begin with a decimal digit, and that is converted
  ;; into a number which is returned as the response class.
  ;; If the string doesn't begin with a decimal digit then the
  ;; response class is -1.
  ;;
  (flet ((match-chars (string pos1 pos2 count)
	   ;; like strncmp
	   (dotimes (i count t)
	     (if* (not (eq (aref string (+ pos1 i))
			   (aref string (+ pos2 i))))
		then (return nil)))))

    (let ((res (make-array 20 :element-type 'character
			   :adjustable t
			   :fill-pointer 0)))
      (if* (null (read-a-line stream res))
	 then ; eof encountered before end of line
	      (return-from wait-for-response (values -1 res)))

      ;; a multi-line response begins with line containing
      ;; a hyphen in the 4th column:
      ;; xyz-  some text
      ;;
      ;;  and ends with a line containing the same reply code but no
      ;;  hyphen.
      ;; xyz  some text
      ;;

      (if* (and (>= (length res) 4) (eq #\- (aref res 3)))
	 then ;; multi line response
	      (let ((old-length (length res))
		    (new-length nil))
		(loop
		  (if* (null (read-a-line stream res))
		     then ; eof encountered before end of line
			  (return-from wait-for-response (values -1 res)))
		  (setq new-length (length res))
		  ;; see if this is the last line
		  (if* (and (>= (- new-length old-length) 4)
			    (eq (aref res (+ old-length 3)) #\space)
			    (match-chars res 0 old-length 3))
		     then (return))

		  (setq old-length new-length))))

      ;; complete response is in res
      ;; compute class and return the whole thing
      (let ((class (or (and (> (length res) 0)
			    (digit-char-p (aref res 0)))
		       -1)))
	(values class res)))))


(defun read-a-line (stream res)
  ;; read from stream and put the result in the adjust able array res
  ;; if line ends in cr-lf, only put a newline in res.
  ;; If we get an eof before the line finishes, return nil,
  ;; else return t if all is ok
  (let (ch last-ch)
    (loop
      (setq ch (read-char stream nil nil))
      (if* (null ch)
	 then ; premature eof
	      (return nil))

      (if* *ftp-debug*
	 then (format *ftp-debug* "~c" ch))

      (if* (eq last-ch #\return)
	 then (if* (eq ch #\linefeed)
		 then (vector-push-extend #\newline res)
		      (return t)
		 else (vector-push-extend last-ch res))
       elseif (eq ch #\linefeed)
	 then ; line ends with just lf, not cr-lf
	      (vector-push-extend #\newline res)
	      (return t)
       elseif (not (eq ch #\return))
	 then (vector-push-extend ch res))

      (setq last-ch ch))))



(defun ftp-command (stream &rest format-args)
  ;; send a command to the ftp server
  (let ((command (apply #'format nil format-args)))
    (if* *ftp-debug*
       then (format *ftp-debug* "to ftp command: ~s~%" command))
    (write-string command stream)
    (write-char #\return stream)
    (write-char #\newline stream)
    (force-output stream)))

































