;; server.cl
;;
;;  
;;  An example of using the acl socket code to create a framework
;; for building lisp server applications.
;;
;;


(in-package :user)



(defvar *server-count* 0)	; used to name servers

(defun make-socket-server (&key (name (format nil "socket server ~d "
					      (incf *server-count*)))
				port 
				function 
				wait 
				(format :text))
  ;;
  ;; create a server process with the given name, listening on the
  ;; given port, running the given function on each connection that
  ;; comes in, and possibly waiting for that function's completion before
  ;; accepting a new connection.
  ;;
  ;; name - a string naming the server process -- if nil, then this 
  ;;	function will create a name.
  ;; port - if nil then an internet domain port number will be chosen
  ;;	    by the operating system.   If a number is given then that
  ;;	    port will be used (or an error will be signalled if it
  ;;	    is already in use).    If port is a string then a unix
  ;;	    domain port will be used.  (this will not work on Windows).
  ;; function - the function to run when a connection is made.  This
  ;;	    function must take one argument which is the stream used
  ;;	    used for reading from and writing to the process that connected
  ;;	    to this socket. 
  ;; wait - if true, then the function will be run in the server process
  ;;	    and thus the server won't accept a new connection until
  ;;	    the function finishes.
  ;; format  - :text (the default) or :binary.   This determes what kind
  ;;	    of data can sent to and read from the socket stream.
  ;;	    
  ;;
  ;;
  ;; The return value is the port number on which the server is
  ;;	listening.
  ;;
  (let ((passive-socket (socket:make-socket
			 :connect :passive
			 :local-port port
			 :address-family 
			 (if* (stringp port)
			    then :file
			  elseif (or (null port)
				     (integerp port))
			    then :internet
			    else (error "illegal value for port: ~s"
					port))
			 :format format
			 )))
    (mp::process-run-function name 
			      #'(lambda ()
				  (start-socket-server passive-socket
						       :function function
						       :wait wait)))
    (socket:local-port passive-socket)))

(defun start-socket-server (passive-socket &key function wait)
  ;; internal function run in the server lightweight process 
  ;; that continually processes the connection.
  ;; This code is careful to ensure that the sockets are 
  ;; properly closed something abnormal happens.
  (unwind-protect
      (loop (let ((connection (socket:accept-connection passive-socket)))
	      (if* wait
		 then (unwind-protect
			  (funcall function connection)
			(errorset (close connection) nil))
		 else (mp:process-run-function
		       (format nil "server run ~d" (incf *server-count*))
		       #'(lambda (con)
			   (unwind-protect
			       (funcall function con)
			     (errorset (close con) nil)))
		       connection))))
    (errorset (close passive-socket) nil)))





(defun room-server (&optional port)
  ;; A sample use of the server code above.
  ;; This creates a server in lisp which, when contacted, will
  ;; send to the contacter the room information about lisp.
  ;; Note that on Unix, the telnet program takes a second argument
  ;; which is the port on which to connect.  Thus after starting
  ;; this serfver up you can 
  ;;  % telnet machine port-number
  ;; and see the room info printed.
  ;;	
  ;; 
  (make-socket-server :name "room server" 
		      :port port
		      :function #'(lambda (port)
				    (let ((*standard-output* port))
				      (room t)))
		      ))



		      
	
(defun eval-server (&optional port)
  ;; 
  ;; create a eval server
  ;; when this server is started, processes can connect to 
  ;; the eval server and it will see a lisp> prompt.
  ;; At this point it can enter an expression and it will
  ;; be read and evaluated by lisp, and the result will be printed.
  ;;
  ;; The easiest way to use this eval server is to use telnet
  ;; to access it.
  ;;
  (make-socket-server :name "eval server"
		      :port port
		      :function #'(lambda (port)
				    (loop 
				      (setq *p* port)
				      (format port "~%lisp> ")
				      (force-output port)
				      (multiple-value-bind (ok value)
					  (errorset (read port) nil)
					(if* (not ok)
					   then (return))
				      
					(multiple-value-bind (ok result)
					    (errorset (eval value) nil)
					  (if* (not ok)
					     then (format port "~%error!~%")
					     else (print result port))))))))


				      
		
    
				
