;; -*- mode: common-lisp; package: excl -*-
;;
;; copyright (c) 1986-1998 Franz Inc, Berkeley, CA  - All rights reserved.
;;
;; Allegro Common Lisp
;; The application-customizable lisp startup code.
;;
;; This code is provided to licensed users of Allegro CL as an example
;; and for restricted reuse.
;;
;; Restricted Rights Legend
;; ------------------------
;; Use, duplication, and disclosure of the software, data and information
;; contained herein by any agency, department or entity of the U.S.
;; Government are subject to restrictions of Restricted Rights for
;; Commercial Software developed at private expense as specified in
;; DOD FAR Supplement 52.227-7013 (c) (1) (ii), as applicable.
;;
;; $Id: aclstart.cl,v 2.48 1998/03/17 23:29:05 duane Exp $

;; This file contains the important lisp code that starts or restarts execution
;; of a lisp image.  This file is distributed in source form so application
;; writers can first understand the mechanisms, and then to customize them as
;; may sometimes be necessary for application delivery.
;;
;; Note carefully the restriction below concerning the code that prints
;; the Franz Inc copyright banner.

(in-package :excl)

;; This mechanism is used for emergency exit e.g. when *terminal-io* is broken.
;; Beware: If this hook is set to '_exit, a process-reset of the initial
;; lisp listener will cause lisp to exit silently and immediately!

(defvar *initial-lisp-listener-exit-hook* 'nil)

;; This function is wrapped around the start of every stack-group computation
;; automatically by the low-level stack-group startup code.
;; User code need not be concerned with it and should not modify it.
;; The source is included here to document the dynamic environment that is
;; automatically wrapped around every computation.

(defun setup-required-stack-group-bindings (continuation &aux vars vals)
  (catch 'sys::internal-stack-group-reset ; Used by process-reset.
    (dolist (b required-stack-group-bindings)
      (unless (member (car b) vars :test #'eq)
	(push (car b) vars)
	(push (eval (cdr b)) vals)))
    (catch 'tpl::top-level-break-loop	; Used for pop etc. from
					; non-lisp-listener stack-group.
      (catch :top-level-reset		; Used for reset from non-lisp-listener
					; stack-group, or restart-function.
	;; Error handlers need to find these variables on every stack group.
	;; Don't make this list longer than it needs to be.
	(progv vars vals
	  (setq vars nil vals nil)
	  ;; Interrupts are disabled until this point in both a dumplisp
	  ;; restart and in a stack-group initial call so that no
	  ;; stack-group reset can happen until inside the catcher.
	  ;; bug1278 - smh 1jun90.
	  ;; The test for #'do-the-dump is a crock designed to treat dumplisp
	  ;; specially.  See .dumplisp in code/misc, and
	  ;; see bug2893 for more details.  - smh 14jan93.
	  (unless (and sys:*current-stack-group*
		       (eq #-os-threads
		           (mp::stack-group-preset-function
			    sys:*current-stack-group*)
			   #+os-threads
		           (mp::thread-preset-function
			    sys:*current-stack-group*)
			   #'do-the-dump)) ;[bug2893]
	    (setq *without-interrupts* nil))
	  (funcall continuation))))))

;; This variable controls whether lisp startup tries to process .clinit files.
;; Upon startup its value is copied to the internal variable, which may be
;; altered by the standard command-line switch processing, see below.
(defvar *read-init-files* t)
(defvar *internal-read-init-files* t)

;; If this is non-NIL it is a function of zero arguments that will be called
;; before tpl:start-interactive-top-level.  It should be used to make any
;; initializations before the top-level.

(defvar *restart-init-function* nil)

;; If this is non-NIL it is a function of zero arguments that will be called
;; to start lisp execution instead of the regular read-eval-print loop.
;; The intention is that it would be the top level of a delivered application.
;; It should never return; if it does, results are unpredictable.
;; During field maintenance and debugging it may be convenient to override
;; the restart function so that the regular read-eval-print loop is entered
;; at startup.  See the treatment of ACL_STARTUP_HOOK below.  Using it, you can
;; execute your application thusly, using csh syntax in this example:
;;  (setenv ACL_STARTUP_HOOK='(setq excl:*restart-app-function* nil)'; yourapp)

(defvar *restart-app-function* nil)

;; This is a list of functions of zero arguments that will be called when a
;; dumplisp image restarts.  These allow the restarting image to do necessary
;; cleanups.  This list is not altered by startup and remains in effect when
;; restarting a subsequent dumplisp, but the function immediately following
;; might be useful to add an action that runs just once.

(defvar *restart-actions* nil)

;; on the windows platform, this is the name of the imagefile we are running
(defvar *image-file* nil)

(eval-when (compile) (declaim (:fbound check-8-bit-input)))

(defun start-lisp-execution ()
  #+microsoft
  (sys::set-stack-cushion #.(* 2 1024 1024))
  #-microsoft
  (sys::set-stack-cushion #.(* 1 1024 1024))

  (setq *clear-input-on-error*
    #-microsoft t
    ;; Make sure this gets set appropriately for build.exe (ie, set to nil):
    #+microsoft (ms-windows-based-app-p))

  (flush-all-logical-pathname-translations)

  (let ((s *initial-terminal-io*))
    ;; Bind all the standard streams to *initial-terminal-io*
    ;; so the startup code can be guaranteed a known I/O environment.

    (handler-case			; Handle any errors during startup.
	(with-standard-io-syntax
	  (let ((*terminal-io* s)
		(*standard-input* s)
		(*standard-output* s)
		(*query-io* s)
		(*debug-io* s)
		(*trace-output* s)
		(*error-output* s)
		(*print-pretty* t))

	    ;; If the scheduler was running in the pre-dumplisp image,
	    ;; it is necessary to clean its state.
	    (when (scheduler-running-p)
	      (funcall-in-package :cleanup-scheduler-state :multiprocessing)
	      (funcall-in-package :kill-processes-after-dumplisp
				  :multiprocessing t))

	    (setq *default-pathname-defaults* (current-directory))

	    ;; You are allowed to suppress (i.e., not display) the Franz
	    ;; Inc. supplied copyright banner in your application, if and
	    ;; only if you agree to display Franz Inc. copyright and
	    ;; proprietary notices wherever you display your own copyright
	    ;; or other proprietary notices.
	    ;;
	    ;; For further information see the :suppress-allegro-cl-banner
	    ;; keyword argument to dumplisp.

	    (when (null .dump-lisp-suppress-allegro-cl-banner.)
	      (copyright-banner))

	    ;; On platforms where foreign functions are loaded (mapped)
	    ;; dynamically, i.e. using .so files, and therefore not
	    ;; written out with the dumplisp image, it is necessary
	    ;; to reload all such libraries when the dumplisp is restarted.
	    ;; This is done with the following call.  It is intended
	    ;; that the path to .so files will be communicated using
	    ;; logical pathnames.  If a .so file cannot be found, an
	    ;; error break will happen which will give the user a
	    ;; chance to supply a new pathname.  The reload has to
	    ;; happen before any calls can be made to a foreign function
	    ;; in a dynamically-loaded .so file. This is as good
	    ;; a place as any to do the reloads, although the following
	    ;; form could be moved slightly earlier or later if a
	    ;; particular application makes additional
	    ;; constaints on the restart procedures, for example, such
	    ;; as needing foreign functions in the earlier
	    ;; kill-processes-after-dumplisp processing, or wants
	    ;; to defer reloading until after command-line processing.

	    (if* (and (excl::featurep :dynload) ; .dll, .sl, or .so style loads
		      (fboundp 'reload-fix-entry-points))
	       then (reload-fix-entry-points (rebornp))

		    ;; On systems that support different lisp runtime
		    ;; kernels, we have to rebuilt all callbacks so
		    ;; they point to the current location of the
		    ;; internal callback machinery
		    ;; The code below is made somewhat complex
		    ;; because we want to avoid loading the foreign
		    ;; function package if it isn't present
		    (if* (and #+ignore (excl::featurep :mswindows)
			      (member (string :foreign-functions)
				      *modules* :test #'string-equal))
		       then (excl::funcall-in-package
			     :rebuild-callback-address-table
			     :foreign-functions)))

	    (setq *internal-read-init-files* *read-init-files*)

	    ;; This will always ignore args on initial build.
	    ;; The command-line processing occurs early so it can
	    ;; change the value of *internal-read-init-file*,
	    ;; *restart-init-function*, or *restart-app-function*.
	    (do-command-line-processing)

	    ;; Will always skip reading on initial build.
	    (read-init-files *internal-read-init-files*)

	    (when *print-startup-message*
	      (print-startup-info t))

	    ;; Warns if *terminal-io* is not supporting 8-bit input.
	    (ics-target-ecase
	     (:+ics (check-8-bit-input))
	     (:-ics nil))

	    ;; Give pre-dumplisp code a chance to run any cleanups.
	    (mapc #'funcall (copy-list *restart-actions*))

	    ;; Fixup the stack-group name and clear the *main-stack-group*.
	    (when sys:*current-stack-group*
	      (setf #-os-threads
	            (mp:stack-group-name sys:*current-stack-group*)
		    #+os-threads
		    (mp::thread-name sys:*current-stack-group*)
		"Initial stack group"))
	    (setq sys::*main-stack-group* nil)

	    ;; call this function to turn on the remote execution
	    ;; interrupt lisp will process remote execution commands
	    #+(or microsoft sun4 sgi4d alpha aix-4-2 hpprism linux86)
	    (excl::sig-handler-remote-execution nil nil)

	    (let ((acl-startup-hook (sys:getenv "ACL_STARTUP_HOOK")))
	      (when acl-startup-hook
		(ignore-errors (eval (read-from-string acl-startup-hook)))))))

      (error (c)
	(ignore-errors
	 (format s
		 (newlinify-format-string
		  "An unhandled error occurred during initialization: ~a~%")
		  c)
	 (force-output s)
	 (exit-lisp -1 :quiet t))))

    (when *restart-init-function*
      (funcall *restart-init-function*))

    #+microsoft
    ;; tell any interested outsider that we're ready for work
    (sig-handler-remote-execution nil nil)

    (tpl:start-interactive-top-level s
				     (or *restart-app-function*
					 #'tpl:top-level-read-eval-print-loop)
				     nil)

    ;; We only get here in exceptional circumstances, namely, if
    ;; someone throws to sys::internal-stack-group-reset.
    ;; This normally happens only if someone
    ;; process-resets out of the initial lisp listener, or more typically, if
    ;; the debugger invokes the error-really-in-troble mechanism.  This can
    ;; happen when lisp's initial terminal-io has been closed remotely,
    ;; so there ;; is little that can be done even to report the lossage.
    ;; This mechanism is implemented in a last ditch effort to make lisp
    ;; go away when it is abandoned in this way.

    ;; Now that we have a better handle on SIGHUP/SIGTERM handling, and
    ;; the ability to associate SIGPIPEs with particular streams, let's
    ;; try less of a panic. In particular, when any sort of socket
    ;; or terminal stream gets a sigpipe, the default action is to
    ;; change-class it to the null stream.  This should
    ;; protect pretty well against recursive SIGPIPE lossage.
    ;; The default handlers for SIGHUP and SIGTERM immediately revert
    ;; the signal ;; action to the default (which is to terminate
    ;; the Unix process) so if a second SIGTERM happens, lisp really
    ;; will go away immediately.  This was a
    ;; problem when running directly in an xterm (and perhaps
    ;; other environments) because when the xterm is told to go away,
    ;; it sends a SIGTERM to lisp, which starts shutting down in an
    ;; orderly way and eventually shuts down its streams.
    ;; But as soon as its streams are closed, the xterm sends _another_
    ;; SIGTERM (or maybe a SIGHUP) which lisp used to try to report.
    ;; But its streams were closed, leading to recursive failure.
    (if* (scheduler-running-p)
       then (when (eq sys::*main-stack-group* sys:*current-stack-group*)
	      (run-hooks *initial-lisp-listener-exit-hook*))
	    (setf #-os-threads
	          (mp::stack-group-state sys:*current-stack-group*)
		  #+os-threads
		  (mp::thread-state sys:*current-stack-group*)
		  :exhausted)
	    ;; Does the following need to be protected against post-dumplisp
	    ;; invalid stack address problems?  I think not... -smh 22jun93
	    (funcall-in-package :internal-stack-group-resume :multiprocessing
				sys::*scheduler-stack-group* nil)
       else ;; Be careful to do no IO in case streams are already shut down
	    ;; or otherwise broken.
	    (exit-lisp -1 :quiet t))
    ))

;; Some typical command-line argument processing:
;;  -q		don't load .clinit.cl in home directory
;;  -qq		don't load any .clinit.cl file.
;;  -e		(EVAL (READ)) of next arg.
;;  -batch		exit on any entry to the debugger.
;;  -backtrace-on-error
;;			do an auto zoom when an error occurs.

;; Remember that the emacs-lisp interface is typically started by a "-e"
;; argument, and this may call (use-background-streams).  Don't depend on
;; initial values of the standard stream variables at startup.

(defun do-command-line-processing ()
  (if* (null sys::.ignore-command-line-arguments.)
     then (do* ((xx (cdr sys::*command-line-arguments*) (cdr xx))
		(x (car xx)  (car xx))
		(y (cadr xx) (cadr xx)))
	      ((or (null xx)
		   (string= "--" x)))
	    (cond ((string= "-e" x)
		   (if* y
		      then (handler-case
			       (eval (with-standard-io-syntax
				       (read-from-string y)))
			     (error (c)
			       (warn "~
An error occurred (~a) during the reading or evaluation of -e ~s" c y)))
			   (setq xx (cdr xx))
		      else (warn "missing companion argument to -e")))
		  ((string= "-f" x)
		   (if* y
		      then (handler-case
			       (funcall (with-standard-io-syntax
					  (read-from-string y)))
			     (error (c)
			       (warn "~
An error occurred (~a) during the reading or evaluation of -f ~s" c y)))
			   (setq xx (cdr xx))
		      else (warn "missing companion argument to -f")))
		  ((string= "" x))
		  ((string= "-C" x)
		   (if* y
		      then (with-standard-io-syntax
			     (compile-file y))
			   (setq xx (cdr xx))
		      else (warn "missing companion argument to -C")))
		  ((string= "-L" x)
		   (if* y
		      then (handler-case
			       (with-standard-io-syntax
				 (load y :verbose nil :autoload t))
			     (error (c)
			       (warn
				"An error occurred (~a) during the load ~s"
				c y)))
			   (setq xx (cdr xx))
		      else (warn "missing companion argument to -L")))
		  ((string= "-q" x)
		   (when *internal-read-init-files*
		     (setq *internal-read-init-files* :no-home)))
		  ((string= "-qq" x)
		   ;; This variable is reset nil each time lisp restarts.
		   (setq *internal-read-init-files* nil))
		  ((string= "-backtrace-on-error" x)
		   (setq *backtrace-on-error* t)
		   (setq *break-hook* #'exit-on-error-hook))
		  ((string= "-batch" x)
		   (setq *batch-mode* t)
		   (setq *break-hook* #'exit-on-error-hook))
		  ((string= "-kill" x)
		   (exit 0))
		  ((string= "-W" x)
		   (setq *break-on-warnings* t))
		  ((string= "-as" x)
		   (pop xx))       ;; the AllegroStore shared library name
		  ((or (string= "-I" x)
		       (string= "-Im" x))
		   (setq *image-file* y)
		   (pop xx))
		  ((string= "-d" x)
		   (dribble y)
		   (pop xx))
		  ((or (string= "-cl" x) ;; the concurrency-limit argument
		       (string= "-H" x) ;; home location
		       )
		   (setq xx (cdr xx)) ;; skip over the associated value
		   )
		  (t (warn "Unknown command-line option ~s" x)))))
  (setf sys::.ignore-command-line-arguments. nil))


;; Support for reading startup initialization files.

;; The name of the file read by the initial top-level upon startup.  It is
;; a list of files.  Since ".clinit.cl" is not an valid ISO 9660 filename
;; (used on CD-ROMs) allow "clinit.cl", too.  Only the first one, in a
;; given directory, is loaded.

(defvar *init-file-names* '(".clinit.cl" "clinit.cl"))

(defun read-init-files (read-init-files)
  (when (or (null read-init-files)
	    ;; Don't process command line args on initial build.
	    (eql *dumplisp-generation-tick* 0))
    (return-from read-init-files nil))
  (let ((site-init (ignore-errors (probe-file "sys:siteinit.cl"))))
    (when site-init
      (load-top-level-init-file site-init :verbose t)))
  (let ((current-pathname (current-directory))
	(u-hd-pathname (multiple-value-bind (success pn)
			   (errorset (user-homedir-pathname) t)
			 (when success pn))))
    ;; save the current pathname [which is the current directory
    ;; upon lisp startup] because it may change when the init file is read.
    ;; Load the user's home init file, then the one in the current directory.
    (if* u-hd-pathname
       then (if* (not (eq :no-home read-init-files))
	       then (load-top-level-init-file-from-directory u-hd-pathname)
		    ;; Make sure that we are not in the user's home directory,
		    ;; because we don't want to load the same file twice.
		    ;; Symbolic links on UNIX make this a bit tricky,
		    ;; so device/inode comparison is necessary.
		    ;; filesys-inode returns as two values the inode
		    ;; and dev numbers returned by Unix stat().
		    (and #+microsoft
			 (not
			  (namestring-compare
			   (namestring (truename u-hd-pathname))
			   (namestring (truename current-pathname))))
			 #-microsoft
			 (not
			  (equal (multiple-value-list
				     (filesys-inode
				      (namestring (truename u-hd-pathname))))
				 (multiple-value-list
				     (filesys-inode
				      (namestring
				       (truename current-pathname))))))
			 (load-top-level-init-file-from-directory
			  current-pathname))
	       else (load-top-level-init-file-from-directory current-pathname))
       else (warn "Skipping attempt to load any initialization files."))))

(defun load-top-level-init-file-from-directory (pathname)
  (dolist (init-file *init-file-names*)
    (setq init-file (merge-pathnames init-file pathname))
    (when (probe-file init-file)
      (load-top-level-init-file init-file)
      (return))))

(defun load-top-level-init-file (pathname &key verbose)
  (multiple-value-bind (ret condition)
      (ignore-errors (load pathname :verbose verbose :if-does-not-exist nil) t)
    (unless ret
      (format t (newlinify-format-string "~&Error loading ~s: ~a~%")
	      pathname condition))))

(defvar .saved-break-level. nil)

(defun exit (&optional (code 0) &rest args &key no-unwind quiet)
  "This exits to the shell or OS, without querying the user, returning
STATUS to the program which invoked Common Lisp.  STATUS must be an
integer whose range is determined by the Operating System, and hence
the value may be truncated for `large' integers.

If NO-UNWIND is nil (the default), then `exit' executes all
outstanding unwind-protect cleanup forms for every process (see
mp:process-kill). If an error occurs during the execution of an
unwind-protect cleanup form or the process becomes inactive, then the
function `exit' could hang forever, waiting for the process to die.
The `exit' function prints the name of the process on which it is
waiting.  Processes which cause `exit' to hang indefinitely (because
the error cause the process to enter a break level) can be killed
manually from another window or editor buffer, or the `exit' can be
interrupted with ^C.  Note, if `exit' is executed again after being
interrupted, the unwind-protect which caused `exit' to hang will not
do so again, because the unwind-protect cleanup code is not
protected (unless it too contains an unwind-protect); for this reason,
`exit' could possibly be re-evaluated.

If NO-UNWIND is non-nil, then `exit' does not evaluate unwind-protect
cleanup forms."
  (declare (ignore no-unwind))
  ;; As soon as we commit to exiting, revert SIGHUP and SIGTERM to the
  ;; default behavior to avoid loops.  If someone kicks us while we are
  ;; trying to leave, it's their fault if we don't close the door.
  (handler-case (when *dribble-active* (dribble))
    (error () nil))
  (unix-signal  1 0)
  (unix-signal 15 0)
  (flet ((exit-1 (&key no-unwind quiet)
	   (if* (boundp 'tpl::*break-level*)
	      then (setq .saved-break-level. tpl::*break-level*))
	   (if* (not (fixnump code)) then (setq code -1))

	   (let ((cleanup sys:*exit-cleanup-forms*))
	     ;; Clear the variable so if we get into trouble, exit will
	     ;; work the second time.
	     (setq sys:*exit-cleanup-forms* nil)
	     (ignore-errors (dolist (x cleanup) (eval x))))

	   (if*  (scheduler-running-p)
	      then ;; This maybe unwinds and kills all but the current process.
		   (funcall-in-package :kill-processes-for-exit
				       :multiprocessing code no-unwind quiet))

	   (if* (not quiet) then  (force-output *terminal-io*))
	   (if* (not no-unwind)
	      then ;; there is only one thread with a catch for non-mp-exit
		   ;; and that is the initial thread.  Thus we catch
		   ;; the case where that throw fails and just do an exit.
		   (handler-case (throw 'non-mp-exit (values code quiet))
		     (condition () (exit-lisp code :quiet quiet))))

	   (if* (not quiet) then (force-output *terminal-io*))
	   (exit-lisp code :quiet quiet)))
    (restart-case (apply #'exit-1 args)
      (nil ()
	  :report "Really exit (:no-unwind t)"
	(exit-1 :no-unwind t :quiet quiet)))))

;; If :quiet is true, exit immediately doing no I/O.  For emergency situations.
(defun exit-lisp (code &key quiet)
  (unless quiet
    (let ((s *initial-terminal-io*))
      (format s "; Exiting Lisp~%")
      #+ignore
      (let ((*terminal-io* s)
	    (*standard-output* s)
	    (tpl::*zoom-print-level* 10)
	    (tpl::*zoom-print-length* 10))
	(tpl:do-command "zoom"
	  :from-read-eval-print-loop nil
	  :count t :all t))
      (force-output s)))
  (excl::.primcall 'sys::lispexit
		   #+microsoft
		   (if* (and .saved-break-level.
			     (= code 0)
			     (> .saved-break-level. 0))
		      then .saved-break-level.
		      else code)
		   #-microsoft code)
  #+ignore
  (let ((s *initial-terminal-io*))
    (format s "; Failed Lisp Exit...~%")
      (force-output s)))
